# -*- coding: utf-8 -*-

from django.core.management import BaseCommand

from hattori import constants
from hattori.exceptions import HattoriException
from hattori.utils import setting, autodiscover_module, get_app_anonymizers


class Command(BaseCommand):
    help = 'This command anonymize real (user-)data of model instances in your database with mock data.'
    modules = None  # List of anonymizers modules. They can be placed in every app
    anonymize_enabled = setting('ANONYMIZE_ENABLED', False)

    def add_arguments(self, parser):
        parser.add_argument(
            '-a',
            '--app',
            help='Only anonymize the given app',
            dest="app",
            metavar="APP"
        )
        parser.add_argument(
            "-m",
            "--model",
            "--models",
            dest="models",
            help="Models to anonymize. Separate multiples by comma.",
            metavar="MODEL"
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            dest="batch_size",
            help="batch size used in the bulk_update of the instances. Depends on the DB machine, defaults use 500.",
            metavar="BATCH_SIZE",
            type=int,
            default=constants.DEFAULT_CHUNK_SIZE,
        )

    def handle(self, *args, **options):
        if not self.anonymize_enabled:
            self.stdout.write(self.style.WARNING('You must set ANONYMIZE_ENABLED to True'))
            exit()

        try:
            modules = autodiscover_module(constants.ANONYMIZER_MODULE_NAME, app_name=options.get('app'))
        except HattoriException as e:
            self.stdout.write(self.style.ERROR(e))
            exit(1)

        total_replacements_count = 0

        for module in modules:
            anonymizers = get_app_anonymizers(module, selected_models=options.get('models'))
            if len(anonymizers) == 0:
                continue

            for anonymizer_class_name in anonymizers:
                anonymizer = getattr(module, anonymizer_class_name)()
                # Start the anonymizing process
                self.stdout.write('{}.{}:'.format(module.__package__, anonymizer.model.__name__))
                num_fields, num_instances, total = anonymizer.run(options.get('batch_size'))
                self.stdout.write(
                    '- {} fields, {} model instances, {} total replacements'.format(num_fields, num_instances, total)
                )
                total_replacements_count += total
        self.stdout.write(self.style.SUCCESS('DONE. Replaced {} values in total'.format(total_replacements_count)))



# -*- coding: utf-8 -*-
class HattoriException(Exception):
    """
    An exception class for Hattori errors
    """
    def __init__(self, msg, original_exception=''):
        super().__init__(msg + (": %s" % original_exception))
        self.original_exception = original_exception



# -*- coding: utf-8 -*-
import inspect
import logging
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from hattori.base import BaseAnonymizer
from hattori.exceptions import HattoriException

logger = logging.getLogger(__name__)


def setting(name, default=None, strict=False):
    """
    Helper function to get a Django setting by name. If setting doesn't exists
    it can return a default or raise an error if in strict mode.

    :param name: Name of setting
    :type name: str
    :param default: Value if setting is unfound
    :param strict: Define if return default value or raise an error
    :type strict: bool
    :returns: Setting's value
    :raises: django.core.exceptions.ImproperlyConfigured if setting is unfound
             and strict mode
    """
    if strict and not hasattr(settings, name):
        msg = "You must provide settings.%s" % name
        raise ImproperlyConfigured(msg)
    return getattr(settings, name, default)


def autodiscover_module(module_name, app_name=None):
    logger.info('Autodiscovering anonymizers modules ...')
    apps_to_search = [app_name] if app_name else settings.INSTALLED_APPS
    modules = []
    for app in apps_to_search:
        try:
            import_module(app)
        except ImportError as e:
            raise HattoriException('ERROR: Can not find app {}'.format(app), e)
        try:
            modules.append(import_module('%s.%s' % (app, module_name)))
        except ImportError as e:
            if app_name:
                raise HattoriException('ERROR: Can not find module {}'.format(module_name, app), e)
    logger.info('Found anonymizers for {} apps'.format(len(modules)))
    return modules


def get_app_anonymizers(module, selected_models=None):
    logger.info('Autodiscovering Anonymizer classes from {} module...'.format(module.__package__))
    models = None
    if selected_models is not None:
        models = [m.strip() for m in selected_models.split(',')]

    if models:
        clazzes = [m[0] for m in inspect.getmembers(module, inspect.isclass)
                   if BaseAnonymizer in m[1].__bases__ and m[1].model.__name__ in models]
    else:
        clazzes = [m[0] for m in inspect.getmembers(module, inspect.isclass) if BaseAnonymizer in m[1].__bases__]
    if len(clazzes) == 0:
        logger.info('Not found any Anonymizer class from {} module'.format(module.__package__))
    return clazzes



# -*- coding: utf-8 -*-

import logging

from tqdm import tqdm
from six import string_types
from django.conf import settings
from bulk_update.helper import bulk_update
from faker import Faker


logger = logging.getLogger(__name__)

try:
    faker = Faker(settings.LANGUAGE_CODE)
except AttributeError:
    faker = Faker()


class BaseAnonymizer:

    model = None
    attributes = None

    def __init__(self):
        if not self.model or not self.attributes:
            logger.info('ERROR: Your anonymizer is missing the model or attributes definition!')
            exit(1)

    def get_query_set(self):
        """
        You can override this in your Anonymizer.
        :return: QuerySet
        """
        return self.model.objects.all()

    def run(self, batch_size):
        instances = self.get_query_set()
        instances_processed, count_instances, count_fields = self._process_instances(instances)
        bulk_update(instances_processed, update_fields=[attrs[0] for attrs in self.attributes], batch_size=batch_size)
        return len(self.attributes), count_instances, count_fields

    def _process_instances(self, instances):
        count_fields = 0
        count_instances = 0

        progress_bar = tqdm(desc="Processing", total=instances.count())
        for model_instance in instances:
            for field_name, replacer in self.attributes:
                if callable(replacer):
                    replaced_value = self.get_allowed_value(replacer, model_instance, field_name)
                elif isinstance(replacer, string_types):
                    replaced_value = replacer
                else:
                    raise TypeError('Replacers need to be callables or Strings!')
                setattr(model_instance, field_name, replaced_value)
                count_fields += 1
            count_instances += 1
            progress_bar.update(1)
        progress_bar.close()
        return instances, count_instances, count_fields

    @staticmethod
    def get_allowed_value(replacer, model_instance, field_name):
        retval = replacer()
        max_length = model_instance._meta.get_field(field_name).max_length
        if max_length:
            retval = retval[:max_length]
        return retval



# -*- encoding: utf-8 -*-

import os
import re
import codecs

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = codecs.open(os.path.abspath(os.path.join(package, '__init__.py')), encoding='utf-8').read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(1)


def get_author(package):
    """
    Return package author as listed in `__author__` in `init.py`.
    """
    init_py = codecs.open(os.path.abspath(os.path.join(package, '__init__.py')), encoding='utf-8').read()
    return re.search("^__author__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(1)


def get_email(package):
    """
    Return package email as listed in `__email__` in `init.py`.
    """
    init_py = codecs.open(os.path.abspath(os.path.join(package, '__init__.py')), encoding='utf-8').read()
    return re.search("^__email__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(1)


def get_long_description():
    """
    return the long description from README.md file
    :return:
    """
    return codecs.open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8').read()


setup(
    name='django-hattori',
    version=get_version('hattori'),
    author=get_author('hattori'),
    author_email=get_email('hattori'),
    url='https://github.com/APSL/django-hattori',
    packages=find_packages(exclude=['tests*']),
    description='Command to anonymize sensitive data.',
    long_description=get_long_description(),
    install_requires=[
        'Django>=1.8',
        'django-bulk-update>=2.2.0',
        'Faker>=0.8.13',
        'six',
        'tqdm>=4.23.4',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
    include_package_data=True,
    zip_safe=False,
)
